#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <asm/param.h>

#ifdef WNOWAIT
/* CPW: Look at procps. */
void stat2proc(char* S, unsigned long *utime, unsigned long *stime, long *cutime, long *cstime) {
    int num;

    char* tmp = strrchr(S, ')');	/* split into "PID (cmd" and "<rest>" */
    *tmp = '\0';			/* replace trailing ')' with NUL */
    num = sscanf(tmp + 2,			/* skip space after ')' too */
       "%*c "
       "%*d %*d %*d %*d %*d "
       "%*u %*u %*u %*u %*u %lu %lu " /* Last two are utime and stime */
       "%ld %ld %*d %*d %*d %*d " /* First two are cutime and cstime */
       "%*u %*u "
       "%*d "
       "%*u %*u %*u %*u %*u %*u "
       "%*s %*s %*s %*s " /* discard, no RT signals & Linux 2.1 used hex */
       "%*u %*u %*u %*d %*d",
       utime, stime, cutime, cstime);
}
#endif

int main(int argc, char *argv[]) {
	FILE *log;
	int i;
	int child;
	struct timeval tv_start;
	struct timeval tv_end;
	struct rusage rusage;
	int status;
	int exitstatus;
	char command[4096];
	int cmdlen = 0;
	int ret;

	double utime;
	double stime;
	double etime;

	double seconds;

	if (argc < 3) {
		fprintf(stderr, "%s log ...\n", argv[0]);
		exit(1);
	}

	log = fopen(argv[1], "a");
	if (!log) {
		fprintf(stderr, "fopen(%s): %s\n", argv[1], strerror(errno));
		exit(1);
	}

	command[0] = '\0';
	for (i = 2; i < argc; i++) {
		int l = strlen(argv[i]);
		if (strcspn(argv[i], " \"\'\\") != strlen(argv[i])) {
			char *tmp;
			int j, k;

			tmp = (char *)malloc(l * 2);
			for (k = j = 0; j < l; j++) {
				switch(argv[i][j]) {
				case '\\':
				case '\"':
					tmp[k++] = '\\';
				default:
					tmp[k++] = argv[i][j];
					break;
				}
			}
			tmp[k] = '\0';
			cmdlen += snprintf(command + cmdlen, sizeof(command) - cmdlen, "\"%s\"", tmp);
			free(tmp);
		} else {
			cmdlen += snprintf(command + cmdlen, sizeof(command) - cmdlen, "%s", argv[i]);
		}
		if (i < argc - 1) {
			command[cmdlen++] = ' ';
		}
		command[cmdlen] = '\0';
	}
	fprintf(log, "command = %s\n", command);
	fflush(log);

	gettimeofday(&tv_start, NULL);

	switch(child = fork()) {
	case 0:
		fclose(log);
		execvp(argv[2], &argv[2]);
		fprintf(stderr, "execvp(%s): %s\n", argv[2], strerror(errno));
		exit(127);
		break;
	case -1:
		perror("fork()");
		exit(1);
		break;
	}
#ifdef WNOWAIT
	do
	{
		FILE *f;
		char path[4096];
		char temp[4096];
		long dummy;
		unsigned long lutime;
		unsigned long lstime;
		double selfutime;
		double selfstime;
		struct siginfo si;

		while((ret = waitid(P_PID, child, &si, WEXITED|WNOWAIT))) {
			if (ret == -1 && errno != EINTR) {
				perror("wait");
				exit(1);
			}
		}

		snprintf(path, sizeof(path), "/proc/%d/stat", child);
		f = fopen(path, "r");
		if (!f) {
			perror("fopen");
			exit(1);
		}
		memset(temp, '\0', sizeof(temp));
		fgets(temp, sizeof(temp), f);
		temp[sizeof(temp) - 1] = '\0';
		fclose(f);

		stat2proc(temp, &lutime, &lstime, &dummy, &dummy);
		selfutime = (double)((double)lutime/HZ);
		selfstime = (double)((double)lstime/HZ);
		fprintf(log, "selfuser = %.06lf\n", selfutime);
		fprintf(log, "selfsys = %.06lf\n", selfstime);
	}
	while (0);
#endif
	while(wait3(&status, 0, &rusage) != child);
	gettimeofday(&tv_end, NULL);

	if (WIFEXITED(status)) {
		exitstatus = WEXITSTATUS(status);
	} else if (WIFSIGNALED(status)) {
		fprintf(stderr, "Terminated by signal %d\n", WTERMSIG(status));
		exitstatus = 128 + WTERMSIG(status);
	} else {
		fprintf(stderr, "aptime: Process neither exited, or was terminated by a signal!\n");
		exitstatus = 1;
	}

	fprintf(log, "status = %d\n", exitstatus);

	seconds = (double)tv_end.tv_sec - (double)tv_start.tv_sec;
	seconds += (double)((double)((double)tv_end.tv_usec - (double)tv_start.tv_usec) / (double)1000000);
	fprintf(log, "elapsed = %.06lf\n", seconds);

	utime = (double)rusage.ru_utime.tv_sec + ((double)rusage.ru_utime.tv_usec / (double)1000000);
	stime = (double)rusage.ru_stime.tv_sec + ((double)rusage.ru_stime.tv_usec / (double)1000000);
	etime = ((tv_end.tv_sec - tv_start.tv_sec) + ((double)(tv_end.tv_usec - tv_start.tv_usec) / (double)1000000));
	fprintf(log, "user = %.06lf\n", utime);
	fprintf(log, "sys = %.06lf\n", stime);

	exit(exitstatus);
}
